package com.persagy.service.impl;

import cn.hutool.core.date.DateUtil;
import com.alibaba.fastjson.JSONObject;
import com.google.common.collect.Lists;
import com.googlecode.aviator.AviatorEvaluator;
import com.googlecode.aviator.Expression;
import com.persagy.cache.AlarmInfoCache;
import com.persagy.cache.CurrentDataCache;
import com.persagy.client.GroupNettyClient;
import com.persagy.entity.*;
import com.persagy.entity.util.ConditionUtil;
import com.persagy.job.ExpireAlarmQueue;
import com.persagy.repository.AlarmRecordRepository;
import com.persagy.service.AlarmHandleService;
import com.persagy.service.AlarmQuartzService;
import com.persagy.utils.DateUtils;
import com.persagy.utils.StringUtil;
import lombok.extern.slf4j.Slf4j;
import org.apache.commons.lang3.StringUtils;
import org.quartz.JobDataMap;
import org.quartz.SchedulerException;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.stereotype.Service;

import java.time.LocalDate;
import java.time.LocalDateTime;
import java.time.LocalTime;
import java.time.format.DateTimeFormatter;
import java.util.*;

/**
 * @description:报警处理实现类：包含报警产生和报警恢复
 * @author:LuoGuanhyi
 * @company:PersagyTechnologyCo.,Ltd
 * @since:2020/10/15 001511:14
 * @version:V1.0
 **/
@Service
@Slf4j
public class AlarmHandleServiceImpl implements AlarmHandleService {
    @Autowired
    AlarmInfoCache alarmInfoCache;
    @Autowired
    CurrentDataCache currentDataCache;
    @Autowired
    GroupNettyClient groupNettyClient;
    @Autowired
    AlarmQuartzService alarmQuartzService;
    @Autowired
    AlarmRecordRepository alarmRecordRepository;
    /**
     * 集团编码
     */
    @Value("${group.code}")
    public String groupCode;
    /**
     * 系统标识，用于创建人
     */
    @Value("${system.id}")
    public String systemId;

    /**
     * @param msg:
     * @description:处理iot采集数据
     * @exception:
     * @author: LuoGuangyi
     * @company: Persagy Technology Co.,Ltd
     * @return: void
     * @since: 2020/10/27 14:28
     * @version: V1.0
     */
    @Override
    public void handleIOTData(String msg) throws SchedulerException, InterruptedException {
        log.info("接收到采集值：[{}]", msg);
        String[] split = msg.split(";");
        if (split.length % 4 != 0) {
            return;
        }
        //返回的信息点个数
        int nums = (split.length) / 4;
        for (int i = 0; i < nums; i++) {
            String dataTime = split[i * 4];
            String meterId = split[i * 4 + 1];
            String funcId = split[i * 4 + 2];
            double value = Double.parseDouble(split[i * 4 + 3]);

            if (alarmInfoCache.hasKey(meterId, funcId)) {
                currentDataCache.putCurrentData(meterId, funcId, value);
                List<AlarmDefine> alarmDefines = alarmInfoCache.getAlarmDefinitionIdByMeterFuncId(meterId, funcId);
                for (AlarmDefine alarmDefine : alarmDefines) {
                    Condition condition = alarmDefine.getCondition();
                    List<JSONObject> codeDetail = condition.getInfoCodes();
                    boolean match = codeDetail.stream().allMatch(
                            p -> currentDataCache.hasKey(p.getString("meterId"), p.getString("funcId"))
                    );
                    //报警定义的所有信息点都有采集数值，具备判断条件
                    if (match) {
                        String defineId = AlarmInfoCache.getAlarmDefineId(alarmDefine);
                        String trigger = condition.getTrigger();
                        String end = condition.getEnd();
                        HashMap<String, Object> paramMap = new HashMap<>();
                        for (JSONObject code : codeDetail) {
                            paramMap.put(
                                    code.getString("infoCode"),
                                    currentDataCache.getCurrentData(code.getString("meterId"), code.getString("funcId")));
                        }
                        Expression triggerExp = AviatorEvaluator.compile(trigger, true);
                        Expression endExp = AviatorEvaluator.compile(end, true);
                        Boolean triggerResult = (Boolean) triggerExp.execute(paramMap);
                        Boolean endResult = (Boolean) endExp.execute(paramMap);
                        log.info("defineId: [{}]", defineId);
                        log.info("triggerResult:[{}],endResult:[{}]", triggerResult, endResult);
                        if (triggerResult && endResult) {
                            log.warn("报警触发条件和报警恢复条件同时满足，请检查，报警定义详情【{}】", alarmDefine.toString());
                        }
                        // 使用intern（）方法，确保上锁的是同一个String对象
                        synchronized (defineId.intern()) {
                            //报警产生值满足（这里的满足不考虑报警持续时间）
                            if (triggerResult) {
                                log.info("满足报警条件");
                                log.info("--" + alarmDefine.toString());
                                log.info("--" + JSONObject.toJSONString(paramMap));
                                //报警的时候不考虑报警恢复，因为同时报警和报警恢复是不应该出现的
                                handlerNowDataAlarm(alarmDefine, dataTime, paramMap);
                            } else {
                                log.info("不满足报警条件");
                                log.info("--" + alarmDefine.toString());
                                log.info("--" + JSONObject.toJSONString(paramMap));
                                //当前数据正常
                                handlerNowDataNormal(alarmDefine, dataTime, endResult, paramMap);
                            }
                        }
                    } else {
                        log.warn("部分信息点没有数值:[{}]", codeDetail);
                    }
                }
            }

        }
    }

    /**
     * @description: 根据报警定义唯一标识从缓存或数据库中获取AlarmDefineState
     * @param: defineId 报警定义唯一标识
     * @exception:
     * @author: lixing
     * @company: Persagy Technology Co.,Ltd
     * @since: 2021/1/8 上午11:05
     * @version: V1.0
     */
    private AlarmDefineState getAlarmDefineStateFromCacheOrDb(String defineId) {
        AlarmDefineState alarmDefineState = alarmInfoCache.getAlarmDefineState(defineId);
        if (Objects.isNull(alarmDefineState)) {
            //默认正常报警状态
            alarmDefineState = new AlarmDefineState(defineId);
            Optional<ZktAlarmRecordDO> recordOptional = alarmRecordRepository.findById(defineId);
            if (recordOptional.isPresent()) {
                ZktAlarmRecordDO alarmRecordDO = recordOptional.get();
                //数据库报警状态：1-未处理
                if ("1".equals(alarmRecordDO.getState())) {
                    alarmDefineState.setState(AlarmDefineState.State.NOT_DEAL.getType());
                    alarmDefineState.setAlarmStartTime(alarmRecordDO.getAlarmTime());
                }
            }
        }
        log.info("alarmDefineState: [{}], （ 0-正常 1-报警）", alarmDefineState.getState());
        return alarmDefineState;
    }

    /**
     * @param alarmDefine :报警定义
     * @param dataTime    :IOT采集时间
     * @param endResult   :报警恢复监测结果
     * @param paramMap    报警恢复触发值
     * @description:当前数据正常判断逻辑
     * @exception:
     * @author: LuoGuangyi
     * @company: Persagy Technology Co.,Ltd
     * @return: void
     * @since: 2020/10/21 16:46
     * @version: V1.0
     */
    private void handlerNowDataNormal(
            AlarmDefine alarmDefine, String dataTime,
            Boolean endResult, HashMap<String, Object> paramMap) throws InterruptedException {
        String defineId = AlarmInfoCache.getAlarmDefineId(alarmDefine);
        AlarmDefineState alarmDefineState = getAlarmDefineStateFromCacheOrDb(defineId);

        Condition condition = alarmDefine.getCondition();
        //当前数据正常，报警状态为正常：清空之前的报警计时，重置回默认状态
        if (AlarmDefineState.State.NORMAL.getType().equals(alarmDefineState.getState())) {
            alarmDefineState = new AlarmDefineState(defineId);
        } else if (AlarmDefineState.State.NOT_DEAL.getType().equals(alarmDefineState.getState())) {
            // 报警恢复条件不满足，不处理
            if (!endResult) {
                return;
            }
            /* 判断是否满足报警恢复持续时间 */
            String alarmSuspendStartTime = alarmDefineState.getAlarmSuspendStartTime();
            if (StringUtil.isEmpty(alarmSuspendStartTime)) {
                alarmSuspendStartTime = dataTime;
                //设置开始恢复时间
                alarmDefineState.setAlarmSuspendStartTime(alarmSuspendStartTime);
            }
            // 报警恢复持续时间
            long alarmSuspendLastTime = DateUtils.betweenTwoTimeSecond(alarmSuspendStartTime, dataTime);
            // 设定的报警恢复持续时间
            int uphold = condition.getEndUphold();
            if (alarmSuspendLastTime < uphold) {
                return;
            }

            /* 超过报警恢复设置的持续时间 */
            // 边缘端记录报警恢复消息
            ZktAlarmRecordDO alarmRecordDO = upsertZktAlarmRecordWhenAlarmSuspend(alarmDefine, dataTime, paramMap);

            String alarmId = alarmRecordDO.getAlarmId();
            //报警恢复参数
            AlarmRecord alarmResumeRecord = AlarmRecord.builder()
                    .state(2)
                    .groupCode(groupCode)
                    .projectId(alarmDefine.getProjectId())
                    .endTime(DateUtils.parseDate(dataTime))
                    .endInfo(JSONObject.toJSONString(paramMap))
                    .build();
            log.info("报警恢复持续时间大于设定时间：[{}]>[{}]", alarmSuspendLastTime, uphold);
            //如果有报警ID,直接报警恢复
            if (StringUtils.isNotEmpty(alarmId)) {
                alarmResumeRecord.setId(alarmId);
                NettyMessage<AlarmRecord> nettyMessage = new NettyMessage<>(6, alarmDefine.getProjectId());
                nettyMessage.setContent(Collections.singletonList(alarmResumeRecord));
                //{"id","123", "state":1, "groupCode":"wd", "projectId":"Pj123","alarmSuspendStartTime":"","endInfo":""}
                groupNettyClient.sendMessage(nettyMessage.toString());
                log.error("产生一条报警恢复消息: [{}]", nettyMessage.toString());
            } else {
                log.info("未获取到报警记录id, 3分钟后重试发送报警恢复消息");
                //如果没有报警ID,定时任务再次测试
                JobDataMap jobDataMap = new JobDataMap();
                jobDataMap.put("alarmRecord", alarmRecordDO.toString());
                jobDataMap.put("refire", "0");
                jobDataMap.put("alarmSuspendStartTime", dataTime);
                jobDataMap.put("endInfo", JSONObject.toJSONString(paramMap));
                jobDataMap.put("defineId", defineId);
                //恢复
                jobDataMap.put("state", "2");
                log.info(JSONObject.toJSONString(jobDataMap));
                String jobName = defineId;
                ExpireAlarmMessage em = new ExpireAlarmMessage();
                //过期消息
                em.setType("1");
                em.setStartTime(DateUtil.offsetMinute(new Date(), 3).toJdkDate());
                em.setJobDataMap(jobDataMap);
                em.setJobName(jobName);
                em.setJobGroupName("resume");
                ExpireAlarmQueue.getExpireAlarmMessageQueue().produce(em);
            }

            String jobName = AlarmInfoCache.getAlarmDefineId(alarmDefine);
            ExpireAlarmMessage em = new ExpireAlarmMessage();
            //取消过期消息
            em.setType("2");
            em.setJobName(jobName);
            em.setJobGroupName("expire");
            ExpireAlarmQueue.getExpireAlarmMessageQueue().produce(em);
            //报警恢复，报警状态重置回默认
            alarmDefineState.reset();
        }
        alarmInfoCache.setAlarmState(defineId, alarmDefineState);
    }

    /**
     * @description: 当报警恢复时边缘端保存/更新报警信息
     * @param: alarmDefine
     * @param: dataTime
     * @param: paramMap
     * @return: com.persagy.entity.ZktAlarmRecordDO
     * @exception:
     * @author: lixing
     * @company: Persagy Technology Co.,Ltd
     * @since: 2021/1/13 下午5:11
     * @version: V1.0
     */
    private ZktAlarmRecordDO upsertZktAlarmRecordWhenAlarmSuspend(AlarmDefine alarmDefine, String dataTime, HashMap<String, Object> paramMap) {
        ZktAlarmRecordDO alarmRecordDO = initZktAlarmRecordDO(alarmDefine);
        alarmRecordDO.setState("2");
        alarmRecordDO.setEndTime(DateUtils.parseDate(dataTime));
        alarmRecordDO.setEndInfo(JSONObject.toJSONString(paramMap));
        alarmRecordRepository.save(alarmRecordDO);
        return alarmRecordDO;
    }

    /**
     * @description: 发送报警仍在继续的消息
     * @param: alarmId 报警记录id
     * @param: lastTime 报警持续时间（单位：s）
     * @param: paramMap 当前采集值
     * @param: projectId 项目id
     * @return: void
     * @exception:
     * @author: lixing
     * @company: Persagy Technology Co.,Ltd
     * @since: 2020/12/17 4:40 下午
     * @version: V1.0
     */
    private void sendAlarmContinueMessage(String alarmId, long lastTime, HashMap<String, Object> paramMap, String projectId)
            throws InterruptedException {
        NettyMessage<JSONObject> nettyMessage = new NettyMessage<>(11, projectId);
        JSONObject jsonObject = new JSONObject();
        jsonObject.put("id", alarmId);
        jsonObject.put("lastTime", lastTime);
        jsonObject.put("iotData", paramMap);
        jsonObject.put("groupCode", groupCode);
        nettyMessage.setContent(Lists.newArrayList(jsonObject));
        groupNettyClient.sendMessage(nettyMessage.toString());
    }

    /**
     * @param alarmDefine 报警定义
     * @param dataTime    :IOT数据采集时间
     * @param paramMap    报警触发值
     * @description:处理当前值报警的情况
     * @exception:
     * @author: LuoGuangyi
     * @company: Persagy Technology Co.,Ltd
     * @return: void
     * @since: 2020/10/20 22:31
     * @version: V1.0
     */
    private void handlerNowDataAlarm(
            AlarmDefine alarmDefine, String dataTime,
            HashMap<String, Object> paramMap) throws InterruptedException {
        // 根据defineId获取报警定义状态
        String defineId = AlarmInfoCache.getAlarmDefineId(alarmDefine);
        AlarmDefineState alarmDefineState = getAlarmDefineStateFromCacheOrDb(defineId);

        Condition condition = alarmDefine.getCondition();
        // 判断报警是否在有效期内
        boolean inEffectTime = ConditionUtil.inEffectiveTime(condition, dataTime);

        if (inEffectTime) {
            // 已经产生报警
            if (AlarmDefineState.State.NOT_DEAL.getType().equals(alarmDefineState.getState())) {
                // 获取报警持续时间
                String alarmStartTime = alarmDefineState.getAlarmStartTime();
                long lastTime = DateUtils.betweenTwoTimeSecond(alarmStartTime, dataTime);
                String alarmId = alarmRecordRepository.findById(defineId).orElse(new ZktAlarmRecordDO()).getAlarmId();
                // 如果已经产生了报警记录
                if (StringUtils.isNotBlank(alarmId)) {
                    log.info("报警已经产生，向云端推送报警持续消息");
                    // 发送报警仍在继续的消息
                    sendAlarmContinueMessage(alarmId, lastTime, paramMap, alarmDefine.getProjectId());
                }
            }
            //之前是是正常状态
            if (AlarmDefineState.State.NORMAL.getType().equals(alarmDefineState.getState())) {
                // 报警持续时间
                long alarmLastTime = 0;
                if (StringUtil.isNotEmpty(alarmDefineState.getAlarmStartTime())) {
                    alarmLastTime = DateUtils.betweenTwoTimeSecond(alarmDefineState.getAlarmStartTime(), dataTime);
                } else {
                    //设置开始报警时间
                    alarmDefineState.setAlarmStartTime(dataTime);
                }

                // 获取报警过期时间
                LocalDateTime expireDateTime = getExpireLocalDateTime(dataTime, condition);
                log.info("报警持续时间：[{}]", alarmLastTime);
                log.info("设定的持续时间：[{}]", condition.getTriggerUphold());
                if (alarmLastTime >= condition.getTriggerUphold()) {
                    log.info("报警持续时间大于设定的持续时间，产生一条报警");
                    // 发送创建报警记录的消息
                    sendCreateAlarmRecordMessage(alarmDefine, dataTime, paramMap,
                            DateUtils.localDateTime2Date(expireDateTime));
                    // 边缘端保存报警信息
                    upsertZktAlarmRecordWhenAlarmStart(alarmDefine);
                    // 报警定义状态更新为未处理
                    alarmDefineState.setState(AlarmDefineState.State.NOT_DEAL.getType());
                    //有过期时间，生成报警过期消息
                    if (expireDateTime != null) {
                        //过期时间
                        log.info("产生一条定时过期报警消息, 过期时间为：{}", expireDateTime);
                        // 创建一条过期任务
                        createExpireJob(alarmDefine, alarmDefineState, expireDateTime);
                    }
                }
            }
        } else {
            // 其他条件全部改成默认值（不报警，不过期，报警开始时间和结束时间为空）
            log.info("[{}]不在生效时间，重置报警状态", dataTime);
            log.info("生效时间为：[{}]", condition.getEffectTime());
            alarmDefineState.reset();
        }
        alarmInfoCache.setAlarmState(defineId, alarmDefineState);
    }

    /**
     * @description: 获取报警过期时间
     * @param: triggerTime 报警触发时间
     * @param: condition 报警条件
     * @return: java.time.LocalDateTime
     * @exception:
     * @author: lixing
     * @company: Persagy Technology Co.,Ltd
     * @since: 2021/1/8 下午6:16
     * @version: V1.0
     */
    private LocalDateTime getExpireLocalDateTime(String triggerTime, Condition condition) {
        // 判断过期设置是否生效
        boolean expireSetEffective = ConditionUtil.expireSetEffective(condition);

        if (!expireSetEffective) {
            return null;
        }

        //过期时间
        String expireTime = ConditionUtil.getEffectEndTime(condition);
        LocalTime localTime = LocalTime.parse(expireTime, DateTimeFormatter.ofPattern(DateUtils.sdfTimeNotDate));
        DateTimeFormatter formatter = DateTimeFormatter.ofPattern(DateUtils.sdfDay);
        LocalDate triggerDate = LocalDate.parse(DateUtils.getDate(triggerTime), formatter);
        return LocalDateTime.of(triggerDate, localTime);
    }

    /**
     * @description: 创建过期任务
     * @param: alarmDefine 报警定义
     * @param: alarmDefineState 报警定义状态对象
     * @param: expireDateTime 过期时间
     * @return: void
     * @exception:
     * @author: lixing
     * @company: Persagy Technology Co.,Ltd
     * @since: 2021/1/13 下午3:39
     * @version: V1.0
     */
    private void createExpireJob(AlarmDefine alarmDefine, AlarmDefineState alarmDefineState, LocalDateTime expireDateTime) throws InterruptedException {
        String defineId = AlarmInfoCache.getAlarmDefineId(alarmDefine);
        String alarmId = alarmRecordRepository.findById(defineId).orElse(new ZktAlarmRecordDO()).getAlarmId();
        ZktAlarmRecordDO alarmRecordDO = ZktAlarmRecordDO.builder()
                .alarmTime(alarmDefineState.getAlarmStartTime())
                .effectEndTime(DateUtils.format(expireDateTime))
                .definitionId(defineId)
                .itemCode(alarmDefine.getItemCode())
                .name(alarmDefine.getName())
                .objId(alarmDefine.getObjId())
                //要变成已过期
                .state("3")
                .alarmId(alarmId)
                .build();
        JobDataMap jobDataMap = new JobDataMap();
        jobDataMap.put("alarmRecord", alarmRecordDO.toString());
        jobDataMap.put("refire", "0");
        jobDataMap.put("expireTime", DateUtils.format(expireDateTime));
        jobDataMap.put("defineId", defineId);
        //过期
        jobDataMap.put("state", "3");
        ExpireAlarmMessage em = new ExpireAlarmMessage();
        //过期消息
        em.setType("1");
        em.setStartTime(DateUtils.localDateTime2Date(expireDateTime));
        em.setJobDataMap(jobDataMap);
        em.setJobName(defineId);
        em.setJobGroupName("expire");
        ExpireAlarmQueue.getExpireAlarmMessageQueue().produce(em);
    }

    /**
     * @description: 当报警开始时边缘端保存/更新报警信息
     * @param: alarmDefine 报警定义
     * @return: com.persagy.entity.ZktAlarmRecordDO
     * @exception:
     * @author: lixing
     * @company: Persagy Technology Co.,Ltd
     * @since: 2021/1/8 下午5:55
     * @version: V1.0
     */
    private ZktAlarmRecordDO upsertZktAlarmRecordWhenAlarmStart(AlarmDefine alarmDefine) {
        ZktAlarmRecordDO zktAlarmRecordDO = initZktAlarmRecordDO(alarmDefine);
        zktAlarmRecordDO.setState("1");
        alarmRecordRepository.save(zktAlarmRecordDO);
        return zktAlarmRecordDO;
    }

    /**
     * @description: 初始化边缘端报警记录对象
     * @param: alarmDefine
     * @return: com.persagy.entity.ZktAlarmRecordDO
     * @exception:
     * @author: lixing
     * @company: Persagy Technology Co.,Ltd
     * @since: 2021/1/13 下午5:07
     * @version: V1.0
     */
    private ZktAlarmRecordDO initZktAlarmRecordDO(AlarmDefine alarmDefine) {
        String defineId = AlarmInfoCache.getAlarmDefineId(alarmDefine);
        ZktAlarmRecordDO zktAlarmRecordDO = alarmRecordRepository.findById(defineId).orElse(new ZktAlarmRecordDO());
        zktAlarmRecordDO.setDefinitionId(defineId);
        zktAlarmRecordDO.setObjId(alarmDefine.getObjId());
        zktAlarmRecordDO.setItemCode(alarmDefine.getItemCode());
        zktAlarmRecordDO.setProjectId(alarmDefine.getProjectId());
        return zktAlarmRecordDO;
    }


    /**
     * @description: 发送创建报警记录消息
     * @param: alarmDefine 报警定义
     * @param: triggerTime 触发时间
     * @param: triggerInfo 触发值
     * @param: expireDate 过期时间
     * @return: void
     * @exception:
     * @author: lixing
     * @company: Persagy Technology Co.,Ltd
     * @since: 2021/1/8 下午5:38
     * @version: V1.0
     */
    private void sendCreateAlarmRecordMessage(
            AlarmDefine alarmDefine, String triggerTime,
            HashMap<String, Object> triggerInfo, Date expireDate) throws InterruptedException {
        Condition condition = alarmDefine.getCondition();
        AlarmRecord alarmRecord = AlarmRecord.builder()
                .category(alarmDefine.getCategory())
                .concern(alarmDefine.getConcern())
                .level(alarmDefine.getLevel())
                .projectId(alarmDefine.getProjectId())
                .state(1)
                .triggerTime(DateUtils.parseDate(triggerTime))
                .remark(alarmDefine.getRemark())
                .triggerInfo(JSONObject.toJSONString(triggerInfo))
                .condition(condition.toString())
                .effectEndTime(expireDate)
                .groupCode(groupCode)
                .itemCode(alarmDefine.getItemCode())
                .objId(alarmDefine.getObjId())
                .classCode(alarmDefine.getClassCode())
                .createUser(systemId)
                .build();
        NettyMessage<AlarmRecord> nettyMessage = new NettyMessage<>(5, alarmDefine.getProjectId());
        nettyMessage.setContent(Collections.singletonList(alarmRecord));
        //推送一条报警记录给远端
        groupNettyClient.sendMessage(nettyMessage.toString());
    }
}
